###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
This module contains low-level functionality of the
`physical_validation.kinetic_energy` module. The functions in this module
should generally not be called directly. Please use the high-level
functions from `physical_validation.kinetic energy`.
"""
from __future__ import print_function
from __future__ import division

import scipy.stats as stats
import numpy as np
import multiprocessing as mproc

from ..util import trajectory
from . import plot


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def temperature(kin, ndof, kb=8.314e-3):
    r"""
    Calculates the temperature acccording to the equipartition theorem.

    .. math::
        T(K) = \frac{2K}{N k_B}

    Parameters
    ----------
    kin : float
        Kinetic energy.
    ndof : float
        Number of degrees of freedom.
    kb : float
        Boltzmann constant :math:`k_B`. Default: 8.314e-3 (kJ/mol).

    Returns
    -------
    temperature : float
        Calculated temperature.
    """
    # ndof * kb * T = 2 * kin
    if isclose(ndof, 0):
        return 0
    return 2 * float(kin) / (float(ndof) * float(kb))


def check_mb_ensemble(kin, temp, ndof, alpha, kb=8.314e-3, verbosity=1,
                      screen=False, filename=None, ene_unit=None):
    r"""
    Checks if a kinetic energy trajectory is Maxwell-Boltzmann distributed.

    .. warning: This is a low-level function. Additionally to being less
       user-friendly, there is a higher probability of erroneous and / or
       badly documented behavior due to unexpected inputs. Consider using
       the high-level version based on the SimulationData object. See
       physical_validation.kinetic_energy.check_mb_ensemble for more
       information and full documentation.

    Parameters
    ----------
    kin : array-like
        Kinetic energy snapshots of the system.
    temp : float
        Target temperature of the system. Used to construct the
        Maxwell-Boltzmann distribution.
    ndof : float
        Number of degrees of freedom in the system. Used to construct the
        Maxwell-Boltzmann distribution.
    alpha : float
        Confidence. TODO: Check proper statistical definition.
    kb : float
        Boltzmann constant :math:`k_B`. Default: 8.314e-3 (kJ/mol).
    verbosity : int
        0: Silent.
        1: Print result details.
        2: Print additional information.
        Default: False.
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.
    ene_unit : string
        Energy unit - used for plotting only.

    Returns
    -------
    result : float
        The p value of the test.

    See Also
    --------
    physical_validation.kinetic_energy.check_mb_ensemble : High-level version
    """

    # Discard burn-in period and time-correlated frames
    kin = trajectory.prepare(kin, verbosity=verbosity, name='Kinetic energy')

    kt = kb * temp
    d, p = stats.kstest(kin, 'chi2', (ndof, 0, kt/2))

    do_plot = screen or filename is not None

    if do_plot:
        hist_sim, k_sim = np.histogram(kin, bins=25, normed=True)
        k_sim = (k_sim[1:] + k_sim[:-1])/2

        ana_dist = stats.chi2(df=ndof, scale=kt/2)
        k_ana = np.linspace(ana_dist.ppf(0.0001),
                            ana_dist.ppf(0.9999), 100)
        hist_ana = ana_dist.pdf(k_ana)

        data = [{'x': k_sim,
                 'y': hist_sim * 100,
                 'name': 'Simulation result'},
                {'x': k_ana,
                 'y': hist_ana * 100,
                 'name': 'Maxwell-Boltzmann'}]

        unit = ''
        if ene_unit is not None:
            unit = ' [' + ene_unit + ']'

        plot.plot(data,
                  legend='best',
                  title='Simulation vs. Maxwell-Boltzmann',
                  xlabel='Kinetic energy' + unit,
                  ylabel='Probability [%]',
                  sci_x=True,
                  filename=filename,
                  screen=screen)

    if verbosity > 0:
        message = ('Kolmogorov-Smirnov test result: p = {:g}\n'
                   'Null hypothesis: Kinetic energy is Maxwell-Boltzmann distributed'.format(p))
        if alpha is not None:
            if p >= alpha:
                message += ('\nConfidence alpha = {:f}\n'
                            'Result: Hypothesis stands'.format(alpha))
            elif p < alpha:
                message += ('\nConfidence alpha = {:f}\n'
                            'Result: Hypothesis rejected'.format(alpha))
        print(message)

    return p


def check_equipartition(positions, velocities, masses,
                        molec_idx, molec_nbonds,
                        natoms, nmolecs,
                        ndof_reduction_tra=0, ndof_reduction_rot=0,
                        dtemp=0.1, temp=None, alpha=0.05,
                        molec_groups=None,
                        random_divisions=0, random_groups=2,
                        ndof_molec=None, kin_molec=None,
                        verbosity=2,
                        screen=False, filename=None):
    r"""
    Checks the equipartition of a simulation trajectory.

    .. warning: This is a low-level function. Additionally to being less
       user-friendly, there is a higher probability of erroneous and / or
       badly documented behavior due to unexpected inputs. Consider using
       the high-level version based on the SimulationData object. See
       physical_validation.kinetic_energy.check_equipartition for more
       information and full documentation.

    Parameters
    ----------
    positions : array-like (nframes x natoms x 3)
        3d array containing the positions of all atoms for all frames
    velocities : array-like (nframes x natoms x 3)
        3d array containing the velocities of all atoms for all frames
    masses : array-like (natoms x 1)
        1d array containing the masses of all atoms
    molec_idx : array-like (nmolecs x 1)
        Index of first atom for every molecule
    molec_nbonds : array-like (nmolecs x 1)
        Number of bonds for every molecule
    natoms : int
        Number of atoms in the system
    nmolecs : int
        Number of molecules in the system
    ndof_reduction_tra : int, optional
        Number of center-of-mass translational degrees of freedom to
        remove. Default: 0.
    ndof_reduction_rot : int, optional
        Number of center-of-mass rotational degrees of freedom to remove.
        Default: 0.
    dtemp : float, optional
        Fraction of temperature deviation tolerated between groups.
        Default: 0.05 (5%).
    temp : float, optional
        Target temperature of the simulation. If None, the kinetic
        energies will not be tested for Maxwell-Boltzmann distribution,
        but only compared amongst each others. Default: None.
    alpha : float, optional
        Confidence for Maxwell-Boltzmann test. Default: 0.05 (5%).
    molec_groups : List[array-like] (ngroups x ?), optional
        List of 1d arrays containing molecule indeces defining groups.
        Useful to pre-define groups of molecules (e.g. solute / solvent,
        liquid mixture species, ...). If None, no pre-defined molecule
        groups will be tested. Default: None.

        *Note:* If an empty 1d array is found as last element in the list, the remaining
        molecules are collected in this array. This allows, for example, to only
        specify the solute, and indicate the solvent by giving an empty array.
    random_divisions : int, optional
        Number of random division tests attempted. Default: 0 (random
        division tests off).
    random_groups : int, optional
        Number of groups the system is randomly divided in. Default: 2.
    ndof_molec : List[dict], optional
        Pass in the degrees of freedom per molecule. Slightly increases speed of repeated
        analysis of the same simulation run.
    kin_molec : List[List[dict]], optional
        Pass in the kinetic energy per molecule. Greatly increases speed of repeated
        analysis of the same simulation run.
    verbosity : int, optional
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 2.
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.

    Returns
    -------
    result : int
        Number of equipartition violations. Tune up verbosity for details.
    ndof_molec : List[dict]
        List of the degrees of freedom per molecule. Can be saved to increase speed of
        repeated analysis of the same simulation run.
    kin_molec : List[List[dict]]
        List of the kinetic energy per molecule per frame. Can be saved to increase speed
        of repeated analysis of the same simulation run.

    See Also
    --------
    physical_validation.kinetic_energy.check_equipartition : High-level version

    """

    dict_keys = ['tot', 'tra', 'rni', 'rot', 'int']

    # for each molecule, calculate total / translational / rotational & internal /
    #   rotational / internal degrees of freedom
    #   returns: List[dict] of floats (shape: nmolecs x 5 x 1)
    if ndof_molec is None:
        ndof_molec = calc_ndof(natoms, nmolecs, molec_idx, molec_nbonds,
                               ndof_reduction_tra, ndof_reduction_rot)

    # for each frame, calculate total / translational / rotational & internal /
    #   rotational / internal kinetic energy for each molecule
    if kin_molec is None:
        try:
            with mproc.Pool() as p:
                kin_molec = p.starmap(calc_molec_kinetic_energy,
                                      [(r, v, masses, molec_idx, natoms, nmolecs)
                                       for r, v in zip(positions, velocities)])
        except AttributeError:
            # Parallel execution doesn't work in py2.7 for quite a number of reasons.
            # Attribute error when opening the `with` region is the first error (and
            # an easy one), but by far not the last. So let's just resort to non-parallel
            # execution:
            kin_molec = [calc_molec_kinetic_energy(r, v, masses, molec_idx, natoms, nmolecs)
                         for r, v in zip(positions, velocities)]

    result = []

    # test system-wide tot, tra, rni, rot, int
    if temp is not None:
        # check for Maxwell-Boltzmann distribution of
        # partitioned kinetic energy trajectories
        result.extend(test_mb_dist(kin_molec=kin_molec,
                                   ndof_molec=ndof_molec,
                                   nmolecs=nmolecs,
                                   temp=temp,
                                   alpha=alpha,
                                   dict_keys=dict_keys,
                                   verbosity=verbosity,
                                   screen=screen,
                                   filename=filename))
    else:
        # compare partitioned temperatures to total temperature
        result.extend(test_temp_diff(kin_molec=kin_molec,
                                     ndof_molec=ndof_molec,
                                     nmolecs=nmolecs,
                                     dtemp=dtemp,
                                     dict_keys=dict_keys,
                                     verbosity=verbosity,
                                     screen=screen,
                                     filename=filename))
    # divide in random groups
    for i in range(random_divisions):
        # randomly assign a group index to each molecule
        group_idx = np.random.randint(random_groups, size=nmolecs)
        # create molecule index for each group
        groups = []
        for rg in range(random_groups):
            groups.append(np.arange(nmolecs)[group_idx == rg])
        # test each group separately
        for rg, group in enumerate(groups):
            if verbosity > 0:
                print('Testing randomly divided group {:d}'.format(rg))
            if verbosity > 3:
                print(group)
            if temp is not None:
                result.extend(test_mb_dist(kin_molec, ndof_molec, nmolecs, temp,
                                           alpha, dict_keys, group, verbosity))
            else:
                result.extend(test_temp_diff(kin_molec, ndof_molec, nmolecs,
                                             dtemp, dict_keys, group, verbosity))
        # test groups against each others
        for rg1, group1 in enumerate(groups):
            for group2 in groups[rg1 + 1:]:
                result.extend(test_temp_diff_groups(kin_molec, ndof_molec, nmolecs,
                                                    group1, group2,
                                                    dtemp, dict_keys, verbosity))

    # use predefined group division?
    # if no groups, return
    if not molec_groups:
        return result, ndof_molec, kin_molec
    # is last group empty?
    last_empty = not molec_groups[-1]
    # get rid of empty groups
    molec_groups = [g for g in molec_groups if g]
    # if no groups now (all were empty), return now
    if not molec_groups:
        return result, ndof_molec, kin_molec

    if last_empty:
        # last group is [] -> insert remaining molecules
        combined = []
        for group in molec_groups:
            combined.extend(group)
        molec_groups[-1] = [m for m in range(nmolecs) if m not in combined]

    for mg, group in enumerate(molec_groups):
        if verbosity > 0:
            print('Testing predifined divided group {:d}'.format(mg))
        if verbosity > 3:
            print(group)
        if temp is not None:
            result.extend(test_mb_dist(kin_molec, ndof_molec, nmolecs, temp,
                                       alpha, dict_keys, group, verbosity))
        else:
            result.extend(test_temp_diff(kin_molec, ndof_molec, nmolecs,
                                         dtemp, dict_keys, group, verbosity))
    # test groups against each others
    if len(molec_groups) > 1:
        for rg1, group1 in enumerate(molec_groups):
            for group2 in molec_groups[rg1 + 1:]:
                result.extend(test_temp_diff_groups(kin_molec, ndof_molec, nmolecs,
                                                    group1, group2,
                                                    dtemp, dict_keys, verbosity))

    return result, ndof_molec, kin_molec


def calc_system_ndof(natoms, nmolecs, nbonds,
                     stop_com_tra, stop_com_rot):
    r"""
    Calculates the total / translational / rotational & internal /
    rotational / internal degrees of freedom of the system.

    Parameters
    ----------
    natoms : int
        Total number of atoms in the system
    nmolecs : int
        Total number of molecules in the system
    nbonds : int
        Total number of bonds in the system
    stop_com_tra : bool
        Was the center-of-mass translation removed during the simulation?
    stop_com_rot : bool
        Was the center-of-mass translation removed during the simulation?

    Returns
    -------
    ndof : dict
        Dictionary containing the degrees of freedom.
        Keys: ['tot', 'tra', 'rni', 'rot', 'int']
    """
    # total ndof
    ndof_tot = 3*natoms - nbonds

    # ndof reduction due to COM motion constraining
    if stop_com_tra:
        ndof_tot -= 3
    if stop_com_rot:
        ndof_tot -= 3

    # translational ndof
    ndof_tot_tra = 3*nmolecs
    if stop_com_tra:
        ndof_tot -= 3

    # rotational & internal ndof
    ndof_tot_rni = ndof_tot - ndof_tot_tra

    # rotational ndof
    ndof_tot_rot = 3*nmolecs
    if stop_com_tra:
        ndof_tot -= 3

    # internal ndof
    ndof_tot_int = ndof_tot_rni - ndof_tot_rot

    # return dict
    ndof = {'tot': ndof_tot,
            'tra': ndof_tot_tra,
            'rni': ndof_tot_rni,
            'rot': ndof_tot_rot,
            'int': ndof_tot_int}
    return ndof


def calc_ndof(natoms, nmolecs,
              molec_idx, molec_nbonds,
              ndof_reduction_tra, ndof_reduction_rot):
    r"""
    Calculates the total / translational / rotational & internal /
    rotational / internal degrees of freedom per molecule.

    Parameters
    ----------
    natoms : int
        Total number of atoms in the system
    nmolecs : int
        Total number of molecules in the system
    molec_idx : List[int]
        Index of first atom for every molecule
    molec_nbonds : List[int]
        Number of bonds for every molecule
    ndof_reduction_tra : int
        Number of center-of-mass translational degrees of freedom to
        remove. Default: 0.
    ndof_reduction_rot : int
        Number of center-of-mass rotational degrees of freedom to remove.
        Default: 0.

    Returns
    -------
    ndof_molec : List[dict]
        List of dictionaries containing the degrees of freedom for each molecule
        Keys: ['tot', 'tra', 'rni', 'rot', 'int']
    """
    # check whether there are monoatomic molecules:
    nmono = (np.array(molec_idx[1:]) - np.array(molec_idx[:-1]) == 1).sum()

    # ndof to be deducted per molecule
    # ndof reduction due to COM motion constraining
    ndof_com_tra_pm = ndof_reduction_tra / nmolecs
    ndof_com_rot_pm = ndof_reduction_rot / (nmolecs - nmono)

    ndof_molec = []
    # add last idx to molec_idx to ease looping
    molec_idx = np.append(molec_idx, [natoms])
    # loop over molecules
    for idx_molec, (idx_atm_init, idx_atm_end) in enumerate(zip(molec_idx[:-1], molec_idx[1:])):
        natoms = idx_atm_end - idx_atm_init
        nbonds = molec_nbonds[idx_molec]
        ndof_tot = 3*natoms - nbonds - ndof_com_tra_pm - ndof_com_rot_pm
        ndof_tra = 3 - ndof_com_tra_pm
        ndof_rni = ndof_tot - ndof_tra
        ndof_rot = 3 - ndof_com_rot_pm
        ndof_int = ndof_tot - ndof_tra - ndof_rot
        if isclose(ndof_int, 0, abs_tol=1e-09):
            ndof_int = 0
        if natoms == 1:
            ndof_tot = 3 - ndof_com_tra_pm
            ndof_tra = 3 - ndof_com_tra_pm
            ndof_rni = 0
            ndof_rot = 0
            ndof_int = 0
        ndof_molec.append({'tot': ndof_tot,
                           'tra': ndof_tra,
                           'rni': ndof_rni,
                           'rot': ndof_rot,
                           'int': ndof_int})

    return ndof_molec


def calc_molec_kinetic_energy(pos, vel, masses,
                              molec_idx, natoms, nmolecs):
    r"""
    Calculates the total / translational / rotational & internal /
    rotational / internal kinetic energy per molecule.

    Parameters
    ----------
    pos : nd-array (natoms x 3)
        2d array containing the positions of all atoms
    vel : nd-array (natoms x 3)
        2d array containing the velocities of all atoms
    masses : nd-array (natoms x 1)
        1d array containing the masses of all atoms
    molec_idx : nd-array (nmolecs x 1)
        Index of first atom for every molecule
    natoms : int
        Total number of atoms in the system
    nmolecs : int
        Total number of molecules in the system

    Returns
    -------
    kin : List[dict]
        List of dictionaries containing the kinetic energies for each molecule
        Keys: ['tot', 'tra', 'rni', 'rot', 'int']
    """
    # add last idx to molec_idx to ease looping
    molec_idx = np.append(molec_idx, [natoms])

    # calculate kinetic energy
    kin_tot = np.zeros(nmolecs)
    kin_tra = np.zeros(nmolecs)
    kin_rni = np.zeros(nmolecs)
    kin_rot = np.zeros(nmolecs)
    kin_int = np.zeros(nmolecs)
    # loop over molecules
    for idx_molec, (idx_atm_init, idx_atm_end) in enumerate(zip(molec_idx[:-1], molec_idx[1:])):
        # if monoatomic molecule
        if idx_atm_end == idx_atm_init + 1:
            v = vel[idx_atm_init]
            m = masses[idx_atm_init]
            kin_tot[idx_molec] = .5 * m * np.dot(v, v)
            kin_tra[idx_molec] = kin_tot[idx_molec]
            kin_rni[idx_molec] = 0
            kin_rot[idx_molec] = 0
            kin_int[idx_molec] = 0
            continue
        # compute center of mass position, velocity and total mass
        com_r = np.zeros(3)
        com_v = np.zeros(3)
        com_m = 0
        # loop over atoms in molecule
        for r, v, m in zip(pos[idx_atm_init:idx_atm_end],
                           vel[idx_atm_init:idx_atm_end],
                           masses[idx_atm_init:idx_atm_end]):
            com_r += m*r
            com_v += m*v
            com_m += m

            # total kinetic energy is straightforward
            kin_tot[idx_molec] += .5 * m * np.dot(v, v)

        com_r /= com_m
        com_v /= com_m

        # translational kinetic energy
        kin_tra[idx_molec] = .5 * com_m * np.dot(com_v, com_v)
        # combined rotational and internal kinetic energy
        kin_rni[idx_molec] = kin_tot[idx_molec] - kin_tra[idx_molec]

        # compute tensor of inertia and angular momentum
        inertia = np.zeros((3, 3))
        angular_mom = np.zeros(3)
        # loop over atoms in molecule
        for r, v, m in zip(pos[idx_atm_init:idx_atm_end],
                           vel[idx_atm_init:idx_atm_end],
                           masses[idx_atm_init:idx_atm_end]):
            # relative positions and velocities
            rr = r - com_r
            rv = v - com_v
            rr2 = np.dot(rr, rr)
            # inertia tensor:
            #   (i,i) = m*(r*r - r(i)*r(i))
            #   (i,j) = m*r(i)*r(j) (i != j)
            atm_inertia = -m*np.tensordot(rr, rr, axes=0)
            for i in range(3):
                atm_inertia[i][i] += m*rr2
            inertia += atm_inertia
            # angular momentum: r x p
            angular_mom += m * np.cross(rr, rv)

        # angular velocity of the molecule: inertia^{-1} * angular_mom
        angular_v = np.dot(np.linalg.inv(inertia), angular_mom)

        # test_kin = 0
        # for r, v, m in zip(pos[idx_atm_init:idx_atm_end],
        #                    vel[idx_atm_init:idx_atm_end],
        #                    masses[idx_atm_init:idx_atm_end]):
        #     # relative positions and velocities
        #     rr = r - com_r
        #     rv = v - com_v - np.cross(angular_v, rr)
        #     test_kin += .5 * m * np.dot(rv, rv)
        #
        # print(test_kin)

        kin_rot[idx_molec] = .5 * np.dot(angular_v, angular_mom)
        kin_int[idx_molec] = kin_rni[idx_molec] - kin_rot[idx_molec]

        # end loop over molecules

    return {'tot': kin_tot,
            'tra': kin_tra,
            'rni': kin_rni,
            'rot': kin_rot,
            'int': kin_int}


def group_kinetic_energy(kin_molec, nmolecs, molec_group=None):
    r"""
    Sums up the partitioned kinetic energy for a
    given group or the entire system.

    Parameters
    ----------
    kin_molec : List[dict]
        Partitioned kinetic energies per molecule.
    nmolecs : int
        Total number of molecules in the system.
    molec_group : iterable
        Indeces of the group to be summed up. None defaults to all molecules
        in the system. Default: None.

    Returns
    -------
    kin : dict
        Dictionary of partitioned kinetic energy for the group.
    """
    #
    kin = {'tot': 0, 'tra': 0, 'rni': 0, 'rot': 0, 'int': 0}
    #
    if molec_group is None:
        molec_group = range(nmolecs)
    # loop over molecules
    for idx_molec in molec_group:
        for key in kin.keys():
            kin[key] += kin_molec[key][idx_molec]

    return kin


def group_ndof(ndof_molec, nmolecs, molec_group=None):
    r"""
    Sums up the partitioned degrees of freedom for a
    given group or the entire system.

    Parameters
    ----------
    ndof_molec : List[dict]
        Partitioned degrees of freedom per molecule.
    nmolecs : int
        Total number of molecules in the system.
    molec_group : iterable
        Indeces of the group to be summed up. None defaults to all molecules
        in the system. Default: None.

    Returns
    -------
    ndof : dict
        Dictionary of partitioned degrees of freedom for the group.
    """
    #
    ndof = {'tot': 0, 'tra': 0, 'rni': 0, 'rot': 0, 'int': 0}
    #
    if molec_group is None:
        molec_group = range(nmolecs)
    # loop over molecules
    for idx_molec in molec_group:
        for key in ndof.keys():
            ndof[key] += ndof_molec[idx_molec][key]

    return ndof


def calc_temperatures(kin_molec, ndof_molec, nmolecs, molec_group=None):
    r"""
    Calculates the partitioned temperature for a
    given group or the entire system.

    Parameters
    ----------
    kin_molec : List[dict]
        Partitioned kinetic energies per molecule.
    ndof_molec : List[dict]
        Partitioned degrees of freedom per molecule.
    nmolecs : int
        Total number of molecules in the system.
    molec_group : iterable
        Indeces of the group to be summed up. None defaults to all molecules
        in the system. Default: None.

    Returns
    -------
    temp : dict
        Dictionary of partitioned temperatures for the group.
    """

    kin = group_kinetic_energy(kin_molec, nmolecs, molec_group)
    ndof = group_ndof(ndof_molec, nmolecs, molec_group)

    temp = {}
    for key in kin:
        temp[key] = temperature(kin[key], ndof[key])

    return temp


def test_mb_dist(kin_molec, ndof_molec, nmolecs,
                 temp, alpha, dict_keys, group=None,
                 verbosity=0, screen=False, filename=None,
                 ene_unit=None):
    r"""
    Tests if the partitioned kinetic energy trajectory of a group (or,
    if group is None, of the entire system) are separately Maxwell-Boltzmann
    distributed.

    Parameters
    ----------
    kin_molec : List[List[dict]]
        Partitioned kinetic energies per molecule for every frame.
    ndof_molec : List[dict]
        Partitioned degrees of freedom per molecule.
    nmolecs : int
        Total number of molecules in the system.
    temp : float
        Target temperature of the simulation.
    alpha : float
        Confidence for Maxwell-Boltzmann test.
    dict_keys : List[str]
        List of dictionary keys representing the partitions of the degrees
        of freedom.
    group : iterable
        Indeces of the group to be tested. None defaults to all molecules
        in the system. Default: None.
    verbosity : int
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 0.
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.
    ene_unit : string
        Energy unit - used for plotting only.

    Returns
    -------
    result : List[float]
        p value for every partition
    """
    # save the partitioned kinetic energy trajectories
    group_kin = {key: [] for key in dict_keys}
    ndof = group_ndof(ndof_molec, nmolecs, group)
    # loop over frames
    for k in kin_molec:
        frame_kin = group_kinetic_energy(k, nmolecs, group)
        for key in dict_keys:
            group_kin[key].append(frame_kin[key])

    result = []
    failed = 0
    # test tot, tra, rni, rot, int
    if verbosity > 1:
        print('Testing whether\n'
              '* total,\n'
              '* translational,\n'
              '* rotational & internal,\n'
              '* rotational, and\n'
              '* internal\n'
              'kinetic energies are Maxwell-Boltzmann distributed.')
    elif verbosity > 0:
        print('Testing whether kinetic energies are Maxwell-Boltzmann distributed.')

    for key in dict_keys:
        p = check_mb_ensemble(kin=group_kin[key], temp=temp, ndof=ndof[key],
                              alpha=alpha, verbosity=verbosity > 2,
                              screen=screen, filename=filename+'_'+key,
                              ene_unit=ene_unit)
        result.append(p)
        if alpha is not None and p < alpha:
            failed += 1
        if verbosity > 1 and alpha is not None:
            if p >= alpha:
                print('* {}: passed'.format(key))
            else:
                print('* {}: failed'.format(key))

    if verbosity > 0 and alpha is not None:
        if failed == 0:
            print('-> Passed')
        else:
            print('-> Failed')

    return result


def test_temp_diff(kin_molec, ndof_molec, nmolecs,
                   dtemp, dict_keys, group=None,
                   verbosity=0, screen=False, filename=None,
                   ene_unit=None):
    r"""
    Tests if the partitioned temperatures (averaged over a trajectory)
    of a group (or, if group is None, of the entire system) are within a
    range `dtemp` of the total temperature.

    Parameters
    ----------
    kin_molec : List[List[dict]]
        Partitioned kinetic energies per molecule for every frame.
    ndof_molec : List[dict]
        Partitioned degrees of freedom per molecule.
    nmolecs : int
        Total number of molecules in the system.
    dtemp : float
        Target temperature of the simulation.
    dict_keys : List[str]
        List of dictionary keys representing the partitions of the degrees
        of freedom.
    group : iterable
        Indeces of the group to be tested. None defaults to all molecules
        in the system. Default: None.
    verbosity : int
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 0.
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.
    ene_unit : string
        Energy unit - used for plotting only.

    Returns
    -------
    result : List[float]
        Temperature ratio to the total temperature for every partition.
    """
    # save the partitioned temperature trajectories
    group_temp = {key: [] for key in dict_keys}
    # loop over frames
    for k in kin_molec:
        frame_temp = calc_temperatures(k, ndof_molec, nmolecs, group)
        for key in dict_keys:
            group_temp[key].append(frame_temp[key])
    # average temperature
    group_temp_avg = {}
    for key in dict_keys:
        group_temp_avg[key] = np.mean(group_temp[key])

    failed = 0
    result = []
    if verbosity > 0 and dtemp is not None:
        print('Testing whether temperatures')
        print('  ' + str(dict_keys[1:]))
        print('are within {:.1f}% of temperature'.format(dtemp*100))
        print('  ' + dict_keys[0] + ' (' + str(group_temp_avg[dict_keys[0]]) + ')')
    elif verbosity > 0:
        print('Testing difference between temperatures')
        print('  ' + str(dict_keys[1:]))
        print('and')
        print('  ' + dict_keys[0] + ' (' + str(group_temp_avg[dict_keys[0]]) + ')')

    temp0 = group_temp_avg[dict_keys[0]]
    for key in dict_keys[1:]:
        result.append(group_temp_avg[key] / temp0)
        if dtemp is not None:
            if (1 - dtemp) * temp0 <= group_temp_avg[key] <= (1 + dtemp) * temp0:
                if verbosity > 1:
                    print('* {} ({:f}): passed'.format(key, group_temp_avg[key]))
            else:
                failed += 1
                if verbosity > 1:
                    print('* {} ({:f}): failed'.format(key, group_temp_avg[key]))

    if verbosity > 0 and dtemp is not None:
        if failed == 0:
            print('-> Passed')
        else:
            print('-> Failed')

    do_plot = screen or filename is not None
    if do_plot:
        data = []
        for key in dict_keys:
            t = group_temp[key]
            data.append({'x': range(0, len(t)),
                         'y': t,
                         'name': 'T(' + key + ')'})

        unit = ''
        if ene_unit is not None:
            unit = ' [' + ene_unit + ']'
        plot.plot(data,
                  legend='best',
                  title='Temperature trajectories',
                  xlabel='Frames',
                  ylabel='Temperature' + unit,
                  sci_x=True,
                  filename=filename,
                  screen=screen)

    return result


def test_temp_diff_groups(kin_molec, ndof_molec, nmolecs,
                          group1, group2,
                          dtemp, dict_keys, verbosity=0):
    r"""
    Tests if the partitioned temperatures (averaged over a trajectory)
    of two groups are (individually) within a range `dtemp` of each others.

    Parameters
    ----------
    kin_molec : List[List[dict]]
        Partitioned kinetic energies per molecule for every frame.
    ndof_molec : List[dict]
        Partitioned degrees of freedom per molecule.
    nmolecs : int
        Total number of molecules in the system.
    group1 : iterable
        Indeces of the first group to be compared.
    group2 : iterable
        Indeces of the second group to be compared.
    dtemp : float
        Target temperature of the simulation.
    dict_keys : List[str]
        List of dictionary keys representing the partitions of the degrees
        of freedom.
    verbosity : int
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 0.

    Returns
    -------
    result : List[float]
        Temperature ratio (first group / second group) for every partition.
    """
    # save the partitioned temperature trajectories (group1)
    group1_temp = {key: [] for key in dict_keys}
    # loop over frames
    for k in kin_molec:
        frame_temp = calc_temperatures(k, ndof_molec, nmolecs, group1)
        for key in dict_keys:
            group1_temp[key].append(frame_temp[key])
    # average temperature
    for key in dict_keys:
        group1_temp[key] = np.mean(group1_temp[key])

    # save the partitioned temperature trajectories (group2)
    group2_temp = {key: [] for key in dict_keys}
    # loop over frames
    for k in kin_molec:
        frame_temp = calc_temperatures(k, ndof_molec, nmolecs, group2)
        for key in dict_keys:
            group2_temp[key].append(frame_temp[key])
    # average temperature
    for key in dict_keys:
        group2_temp[key] = np.mean(group2_temp[key])

    result = []
    failed = 0
    if verbosity > 0 and dtemp is not None:
        print('Testing whether temperatures of both groups are within {:.1f}%'.
              format(dtemp*100))

    for key in dict_keys:
        result.append(group1_temp[key]/group2_temp[key])
        if dtemp is not None:
            if (1. - dtemp) <= group1_temp[key]/group2_temp[key] <= (1 + dtemp):
                if verbosity > 1:
                    print('* {}: passed'.format(key))
            else:
                failed += 1
                if verbosity > 1:
                    print('* {}: failed'.format(key))

    if verbosity > 0 and dtemp is not None:
        if failed == 0:
            print('-> Passed')
        else:
            print('-> Failed')

    return result



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structure carrying information on the simulated system.
"""
import warnings
import numpy as np

from ..util import error as pv_error


class SystemData(object):
    r"""SystemData: Information about the atoms and molecules in the system.

    The information stored in SystemData objects describes the atom and molecules
    in the system as far as the physical validation tests need it.

    The system is described in terms of

    * natoms: the total number of atoms in the system
    * nconstraints: the total number of constraints in the system
    * ndof_reduction_tra: global reduction of translational degrees of freedom (e.g.
      due to constraining the center of mass of the system)
    * ndof_reduction_rot: global reduction of rotational degrees of freedom (e.g.
      due to constraining the center of mass of the system)

    The atoms are described in terms of

    * mass: a list of the mass of every atom in the system

    The molecules are described by

    * molecule_idx: a list with the indices first atoms of every molecule (this assumes
      that the atoms are sorted by molecule)
    * nconstraints_per_molecule: a list with the number of constraints in every molecule

    Only used internally:

    * ndof_per_molecule: a list with the number of degrees of freedom of every molecule

    Reserved for future use:

    * bonds
    * constrained_bonds

    *Notes:*

    * kinetic_energy.mb_ensemble() only requires information on the system
      (natoms, nconstraints, ndof_reduction_tra, ndof_reduction_rot)
    * kinetic_energy.equipartition() additionally requires information on the atoms and molecules
      (mass, molecule_idx, nconstraints_per_molecule)

    All other tests do not require and information from SystemData.

    """

    def __init__(self,
                 natoms=None, nconstraints=None,
                 ndof_reduction_tra=None, ndof_reduction_rot=None,
                 mass=None, molecule_idx=None, nconstraints_per_molecule=None):
        self.__natoms = None
        self.__nconstraints = None
        self.__ndof_reduction_tra = None
        self.__ndof_reduction_rot = None
        self.__mass = None
        self.__molecule_idx = None
        self.__nconstraints_per_molecule = None
        self.__ndof_per_molecule = None
        self.__bonds = None
        self.__constrained_bonds = None

        if natoms is not None:
            self.natoms = natoms
        if nconstraints is not None:
            self.nconstraints = nconstraints
        if ndof_reduction_tra is not None:
            self.ndof_reduction_tra = ndof_reduction_tra
        if ndof_reduction_rot is not None:
            self.ndof_reduction_rot = ndof_reduction_rot
        if mass is not None:
            self.mass = mass
        if molecule_idx is not None:
            self.molecule_idx = molecule_idx
        if nconstraints_per_molecule is not None:
            self.nconstraints_per_molecule = nconstraints_per_molecule

    @property
    def natoms(self):
        """int: Number of atoms in the system

        """
        return self.__natoms

    @natoms.setter
    def natoms(self, natoms):
        self.__natoms = int(natoms)

    @property
    def nconstraints(self):
        """float: Total number of constraints in the system

        Does not include the reduction of degrees of freedom in the absence of
        external forces.

        """
        return self.__nconstraints

    @nconstraints.setter
    def nconstraints(self, nconstraints):
        self.__nconstraints = float(nconstraints)

    @property
    def ndof_reduction_tra(self):
        """float: Number of translational degrees of freedom deducted
        from 3*[# of molecules]

        """
        return self.__ndof_reduction_tra

    @ndof_reduction_tra.setter
    def ndof_reduction_tra(self, ndof_reduction_tra):
        self.__ndof_reduction_tra = float(ndof_reduction_tra)

    @property
    def ndof_reduction_rot(self):
        """float: Number of rotational degrees of freedom deducted
        from 3*[# of molecules]

        """
        return self.__ndof_reduction_rot

    @ndof_reduction_rot.setter
    def ndof_reduction_rot(self, ndof_reduction_rot):
        self.__ndof_reduction_rot = float(ndof_reduction_rot)

    @property
    def mass(self):
        """nd-array: Mass vector for the atoms

        Setter accepts array-like objects.

        """
        return self.__mass

    @mass.setter
    def mass(self, mass):
        mass = np.asarray(mass)
        if mass.ndim != 1:
            raise pv_error.InputError('mass',
                                      'Expected 1-dimensional array.')
        if self.natoms is None:
            self.natoms = mass.size
        elif mass.size != self.natoms:
            raise pv_error.InputError('mass',
                                      'Mass vector does not have length == natoms.')
        self.__mass = mass

    @property
    def molecule_idx(self):
        """nd-array: List of index of first atom of each molecule

        Setter accepts array-like objects.

        """
        return self.__molecule_idx

    @molecule_idx.setter
    def molecule_idx(self, molecule_idx):
        molecule_idx = np.asarray(molecule_idx)
        if molecule_idx.ndim != 1:
            raise pv_error.InputError('molecule_idx',
                                      'Expected 1-dimensional array.')
        if (self.nconstraints_per_molecule is not None and
           self.nconstraints_per_molecule.shape != molecule_idx.shape):
            warnings.warn('New `molecule_idx` does not have the same'
                          'shape as previously set `nconstraints_per_molecule`.'
                          'Setting `nconstraints_per_molecule = None` to avoid'
                          'errors.')
            self.__nconstraints_per_molecule = None
        self.__molecule_idx = molecule_idx

    @property
    def nconstraints_per_molecule(self):
        """nd-array: List of number of constraints per molecule

        Setter accepts array-like objects.

        """
        return self.__nconstraints_per_molecule

    @nconstraints_per_molecule.setter
    def nconstraints_per_molecule(self, nconstraints_per_molecule):
        nconstraints_per_molecule = np.array(nconstraints_per_molecule)
        if nconstraints_per_molecule.ndim != 1:
            raise pv_error.InputError('nconstraints_per_molecule',
                                      'Expected 1-dimensional array.')
        if self.molecule_idx is not None:
            if nconstraints_per_molecule.shape != self.molecule_idx.shape:
                raise pv_error.InputError('nconstraints_per_molecule',
                                          'Expected `nconstraints_per_molecule` to have'
                                          'the same shape as `moldecule_idx`.')

        self.__nconstraints_per_molecule = nconstraints_per_molecule

    @property
    def ndof_per_molecule(self):
        """nd-array: List of number of degrees of freedom per molecule

        Setter accepts array-like objects.

        """
        return self.__ndof_per_molecule

    @ndof_per_molecule.setter
    def ndof_per_molecule(self, ndof_per_molecule):
        # used internally - check for consistency?
        self.__ndof_per_molecule = ndof_per_molecule

    @property
    def bonds(self):
        """List[List[int]]: List of bonds per molecule
        """
        return self.__bonds

    @bonds.setter
    def bonds(self, bonds):
        self.__bonds = bonds

    @property
    def constrained_bonds(self):
        """List[List[int]]: List of constrained bonds per molecule
        """
        return self.__constrained_bonds

    @constrained_bonds.setter
    def constrained_bonds(self, constrained_bonds):
        self.__constrained_bonds = constrained_bonds



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################

from __future__ import division

import numpy as np
from scipy import stats

from pymbar import timeseries

from . import error as pv_error


# Pymbar v4 to v3 interface adapter
try:
    timeseries.detectEquilibration
except AttributeError:
    timeseries.detectEquilibration = timeseries.detect_equilibration
try:
    timeseries.subsampleCorrelatedData
except AttributeError:
    timeseries.subsampleCorrelatedData = timeseries.subsample_correlated_data


def equilibrate(traj, verbose=False, name=None):
    traj = np.array(traj)
    if traj.ndim == 1:
        t0, g, n_eff = timeseries.detectEquilibration(traj)
        if t0 == 0 and traj.size > 10:
            # See https://github.com/choderalab/pymbar/issues/277
            t0x, gx, n_effx = timeseries.detectEquilibration(traj[10:])
            if t0x != 0:
                t0 = t0x + 10
        n = traj.size
        res = traj[t0:]

    elif traj.ndim == 2 and traj.shape[0] == 2:
        t01, g1, n_eff1 = timeseries.detectEquilibration(traj[0])
        t02, g2, n_eff2 = timeseries.detectEquilibration(traj[1])
        t0 = max(t01, t02)
        if t0 == 0 and traj.shape[1] > 10:
            # See https://github.com/choderalab/pymbar/issues/277
            t01x, g1x, n_eff1x = timeseries.detectEquilibration(traj[0, 10:])
            t02x, g2x, n_eff2x = timeseries.detectEquilibration(traj[1, 10:])
            t0x = max(t01x, t02x)
            if t0x != 0:
                t0 = t0x + 10
        n = traj.shape[1]
        res = traj[:, t0:]
    elif traj.ndim == 2:
        raise NotImplementedError('trajectory.equilibrate() in 2 dimensions is only '
                                  'implemented for exactly two timeseries.')
    else:
        raise NotImplementedError('trajectory.equilibrate() is not implemented for '
                                  'trajectories with more than 2 dimensions.')

    if verbose:
        if not name:
            name = 'Trajectory'
        if t0 == 0:
            print('{:s} equilibration: No frames discarded for burn-in.'.format(name))
        elif t0 == 1:
            print('{:s} equilibration: First frame ({:.1%} of '
                  'trajectory) discarded for burn-in.'.format(name, 1 / n))
        else:
            print('{:s} equilibration: First {:d} frames ({:.1%} of '
                  'trajectory) discarded for burn-in.'.format(name, t0, t0 / n))

    return res


def decorrelate(traj, facs=None, verbose=False, name=None):
    traj = np.array(traj)
    if traj.ndim == 1:
        idx = timeseries.subsampleCorrelatedData(traj)
        n0 = traj.size
        n1 = len(idx)
        res = traj[idx]
    elif facs is not None:
        # The cleanest way to decorrelate multi-dimensional trajectories would probably
        # be a sort of "parallel-decorrelation", taking frames in a way that both trajectories
        # are independently decorrelated. pymbar does not offer this functionality, so for
        # now, here's a work-around: We'll decorrelate such that
        #     traj_sum = facs[0]*traj[0, :] + facs[1]*traj[1, :] + ...
        # is decorrelated.
        # Use case:
        #     traj_sum = 1.0 * U + P * V
        traj_sum = np.zeros(traj.shape[1])
        for n, f in enumerate(facs):
            traj_sum += f * traj[n]
        idx = timeseries.subsampleCorrelatedData(traj_sum)
        n0 = traj.shape[1]
        n1 = len(idx)
        res = traj[:, idx]
    else:
        raise NotImplementedError('trajectory.decorrelate() is not implemented for '
                                  'trajectories with more than 1 dimension.')
    if verbose:
        n = n0 - n1
        if not name:
            name = 'Trajectory'
        if n == 0:
            print('{:s} decorrelation: No frames discarded for decorrelation.'.format(name))
        elif n == 1:
            print('{:s} decorrelation: 1 frame ({:.1%} of '
                  'trajectory) discarded for decorrelation.'.format(name, 1/n0))
        else:
            print('{:s} decorrelation: {:d} frames ({:.1%} of '
                  'trajectory) discarded for decorrelation.'.format(name, n, n/n0))

    return res


def cut_tails(traj, cut, verbose=False, name=None):
    traj = np.array(traj)
    dc = 100 * cut
    if traj.ndim == 1:
        tmax = stats.scoreatpercentile(traj, 100 - dc)
        tmin = stats.scoreatpercentile(traj, dc)
        t = traj[(tmin <= traj) * (traj <= tmax)]
        n0 = traj.size
        n = t.size
    elif traj.ndim == 2:
        tmax = stats.scoreatpercentile(traj, 100 - dc, axis=1)
        tmin = stats.scoreatpercentile(traj, dc, axis=1)
        t = traj[:,
                 (tmin[0] <= traj[0]) * (tmin[1] <= traj[1]) *
                 (tmax[0] >= traj[0]) * (tmax[1] >= traj[1])]
        n0 = traj.shape[1]
        n = t.shape[1]
    else:
        raise NotImplementedError('trajectory.cut_tails() is not implemented for '
                                  'trajectories with more than 2 dimension.')

    if verbose:
        if not name:
            name = 'Trajectory'
        print('{:s} tails (cut = {:.2%}): {:n} frames ({:.2%} of trajectory) were cut'.format(
              name, cut, n0 - n, (n0-n)/n0))

    return t


def prepare(traj, cut=None, facs=None, verbosity=1, name=None):
    traj = np.array(traj)
    if not name:
        name = 'Trajectory'

    def traj_length(t):
        if t.ndim == 1:
            return t.size
        else:
            return t.shape[1]

    if traj.ndim > 2:
        raise NotImplementedError('trajectory.prepare() is not implemented for '
                                  'trajectories with more than 2 dimensions.')

    # original length
    n0 = traj_length(traj)
    # equilibrate
    res = equilibrate(traj, verbose=False)
    n1 = traj_length(res)
    if verbosity > 2:
        print('{:s} equilibration: First {:d} frames ({:.1%} of '
              'trajectory) discarded for burn-in.'.format(name, n0 - n1, (n0 - n1) / n0))
    # decorrelate
    res = decorrelate(res, facs=facs, verbose=False)
    n2 = traj_length(res)
    if verbosity > 2:
        print('{:s} decorrelation: {:d} frames ({:.1%} of equilibrated '
              'trajectory) discarded for decorrelation.'.format(name, n1 - n2, (n1 - n2)/n1))
    # cut tails
    if cut is not None:
        res = cut_tails(res, cut, verbose=False)
        n3 = traj_length(res)
        if verbosity > 2:
            print('{:s} tails (cut = {:.2%}): {:n} frames ({:.2%} of equilibrated and '
                  'decorrelated trajectory) were cut'.format(name, cut, n2 - n3, (n2 - n3)/n2))
    # end length
    nn = traj_length(res)

    if verbosity > 0:
        print('After equilibration, decorrelation and tail pruning, {:.2%} ({:n} frames) '
              'of original {:s} remain.'.format(nn/n0, nn, name))

    return res


def overlap(traj1, traj2, cut=None, verbose=False, name=None):
    traj1 = np.array(traj1)
    traj2 = np.array(traj2)
    if traj1.ndim == traj2.ndim and traj2.ndim == 1:
        if cut:
            dc = 100 * cut
            max1 = stats.scoreatpercentile(traj1, 100 - dc)
            min1 = stats.scoreatpercentile(traj1, dc)
            max2 = stats.scoreatpercentile(traj2, 100 - dc)
            min2 = stats.scoreatpercentile(traj2, dc)
        else:
            max1 = traj1.max()
            min1 = traj1.min()
            max2 = traj2.max()
            min2 = traj2.min()

        tmin = max(min1, min2)
        tmax = min(max1, max2)

        t1 = traj1[(tmin <= traj1) * (traj1 <= tmax)]
        t2 = traj2[(tmin <= traj2) * (traj2 <= tmax)]
    elif traj1.ndim == traj2.ndim and traj2.ndim == 2:
        if traj1.shape[0] != 2 or traj2.shape[0] != 2:
            raise NotImplementedError('trajectory.overlap() in 2 dimensions is only '
                                      'implemented for exactly two timeseries per trajectory.')
        if cut:
            dc = 100 * cut
            max1 = stats.scoreatpercentile(traj1, 100 - dc, axis=1)
            min1 = stats.scoreatpercentile(traj1, dc, axis=1)
            max2 = stats.scoreatpercentile(traj2, 100 - dc, axis=1)
            min2 = stats.scoreatpercentile(traj2, dc, axis=1)
        else:
            max1 = traj1.max(axis=1)
            min1 = traj1.min(axis=1)
            max2 = traj2.max(axis=1)
            min2 = traj2.min(axis=1)

        tmin = np.max([min1, min2], axis=0)
        tmax = np.min([max1, max2], axis=0)

        t1 = traj1[:,
                   (tmin[0] <= traj1[0]) * (tmin[1] <= traj1[1]) *
                   (tmax[0] >= traj1[0]) * (tmax[1] >= traj1[1])]
        t2 = traj2[:,
                   (tmin[0] <= traj2[0]) * (tmin[1] <= traj2[1]) *
                   (tmax[0] >= traj2[0]) * (tmax[1] >= traj2[1])]
    elif traj1.ndim != traj2.ndim:
        raise pv_error.InputError(['traj1', 'traj2'],
                                  'Trajectories don\'t have the same number of dimensions')
    else:
        raise NotImplementedError('trajectory.overlap() is not implemented for '
                                  'trajectories with more than 2 dimensions.')

    if np.any(max1 < min2) or np.any(max2 < min1):
        if verbose:
            if not name:
                name = 'Trajectory'
            print('{:s} overlap: No overlap found between trajectories'.format(name))
        return np.array([]), np.array([]), None, None

    if verbose:
        if not name:
            name = 'Trajectory'
        print('{:s} overlap: {:.1%} of trajectory 1, and {:.1%} of trajectory 2 '
              'were found within overlap region.\n'
              '              That corresponds to {:n} frames and {:n} frames, '
              'respectively'.format(name, len(traj1)/len(t1), len(traj2)/len(t2),
                                    len(t1), len(t2)))

    return t1, t2, tmin, tmax



from __future__ import print_function, division, absolute_import


import sys
import os
import shutil
import json
import argparse
import re
import math
from collections import OrderedDict

from physical_validation import integrator, ensemble, kinetic_energy
from physical_validation.util.gromacs_interface import GromacsInterface
from physical_validation.data.gromacs_parser import GromacsParser


def mkdir_bk(dirname, verbose=False, nobackup=False):
    if os.path.exists(dirname) and nobackup:
        shutil.rmtree(dirname)
    elif os.path.exists(dirname):
        if verbose:
            print('Directory ' + dirname + ' exists. Backing it up.')
        basename = os.path.basename(dirname)
        n = 1
        bk_dir = '#' + basename + '_' + str(n) + '#'
        while os.path.exists(dirname.replace(basename, bk_dir)):
            n += 1
            bk_dir = '#' + basename + '_' + str(n) + '#'
        os.rename(dirname, dirname.replace(basename, bk_dir))
    os.makedirs(dirname)


def file_bk(filename, verbose=False):
    if os.path.exists(filename):
        if verbose:
            print('File ' + filename + ' exists. Backing it up.')
        basename = os.path.basename(filename)
        n = 1
        bk_file = '#' + basename + '_' + str(n) + '#'
        while os.path.exists(filename.replace(basename, bk_file)):
            n += 1
            bk_file = '#' + basename + '_' + str(n) + '#'
        os.rename(filename, filename.replace(basename, bk_file))


def basic_run_cmds(directory, grompp_args=None, mdrun_args=None):
    grompp = '$GROMPPCMD -f system.mdp -p system.top -c system.gro -o system.tpr'
    if grompp_args:
        for arg in grompp_args:
            grompp += ' ' + arg
    mdrun = '$MDRUNCMD -s system.tpr -deffnm system'
    if mdrun_args:
        for arg in mdrun_args:
            mdrun += ' ' + arg
    return [
        'oldpath=$PWD',
        'cd ' + directory,
        grompp + ' && ' + mdrun,
        'cd $oldpath'
    ]


class Test(object):
    @classmethod
    def parser(cls):
        raise NotImplementedError

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        raise NotImplementedError

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        raise NotImplementedError

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup):
        raise NotImplementedError

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity):
        raise NotImplementedError


class IntegratorTest(Test):
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(
            description='Tests the integrator convergence.',
            formatter_class=argparse.RawTextHelpFormatter,
            prog=cls.__name__
        )
        parser.add_argument('-n', '--n_iterations', type=int, default=2,
                            help='The number of different time steps tested.')
        parser.add_argument('-t', '--tolerance', type=float, default=0.1,
                            help=('The relative tolerance accepted. Default: 0.1.\n'
                                  'Example: By default, the test passes if\n'
                                  '         3.6 <= dE(2*dt)/dE(dt) <= 4.4.')
                            )

        return parser

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        args = cls.parser().parse_args(args)
        return cls.prepare(input_dir, target_dir, system_name, nobackup,
                           n_iterations=args.n_iterations)

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        args = cls.parser().parse_args(args)
        return cls.analyze(gmx_parser, system_dir, system_name, base_data, verbosity,
                           tolerance=args.tolerance, n_iterations=args.n_iterations)

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup, n_iterations=None):
        # Standard value
        if n_iterations is None:
            n_iterations = cls.parser().get_default('n_iterations')
        # Read base options
        options = GromacsInterface.read_mdp(os.path.join(input_dir, 'system.mdp'))

        # Check if options relevant to integrator tests are set
        # Set to standard if not - I am not happy about hardcoding this here...
        # For future: Find a way to obtain standards from GROMACS build
        if 'nsteps' not in options:
            raise ValueError('System ' + system_name + ' has no \'nsteps\' defined in mdp file. ' +
                             'Running integrator test does not make sense.')
        if 'dt' not in options:
            options['dt'] = str(0.001)
        if 'nstcalcenergy' not in options:
            options['nstcalcenergy'] = str(100)
        if 'nstenergy' not in options:
            options['nstenergy'] = str(1000)
        # if 'nstlist' not in options:
        #     options['nstlist'] = str(10)

        # Prepare folders for iterations
        directories = []
        for n in range(1, n_iterations+1):
            current_dir = os.path.join(target_dir, 'integrator_' + str(n))
            mkdir_bk(current_dir, nobackup=nobackup)
            # update timesteps, length and intervals
            options['dt'] = str(float(options['dt'])*0.5)
            for key in ['nsteps', 'nstcalcenergy', 'nstenergy']:  # , 'nstlist']:
                options[key] = str(int(options[key])*2)
            # write mdp
            GromacsInterface.write_mdp(options, os.path.join(current_dir, 'system.mdp'))
            # copy topology and starting structure
            for suffix in ['gro', 'top']:
                shutil.copy2(os.path.join(input_dir, 'system.' + suffix),
                             current_dir)
            directories.append(current_dir)

        return directories

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity,
                tolerance=None, n_iterations=None):
        # Standard value
        if tolerance is None:
            tolerance = cls.parser().get_default('tolerance')
        if n_iterations is None:
            n_iterations = cls.parser().get_default('n_iterations')

        # list of results from simulations at different time steps
        results = []

        # base data
        if base_data['reduced'] is None:
            current_dir = os.path.join(system_dir, 'base')
            base_data['reduced'] = gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.gro')
            )
        results.append(base_data['reduced'])

        # append data at reduced dt
        for n in range(1, n_iterations+1):
            current_dir = os.path.join(system_dir, 'integrator_' + str(n))
            results.append(gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.gro')
            ))

        # run test
        deviation = integrator.convergence(simulations=results,
                                           verbose=(verbosity > 0),
                                           convergence_test='max_deviation')

        result = deviation <= tolerance
        if result:
            message = 'IntegratorTest PASSED (max deviation = {:.2g}, tolerance = {:.2g})'.format(
                deviation, tolerance
            )
        else:
            message = 'IntegratorTest FAILED (max deviation = {:.2g}, tolerance = {:.2g})'.format(
                deviation, tolerance
            )

        return {'test': result,
                'result': deviation,
                'tolerance': tolerance,
                'message': message}


class EnsembleTest(Test):
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(
            description='Tests the validity of the potential energy and / or volume ensemble.',
            formatter_class=argparse.RawTextHelpFormatter,
            prog=cls.__name__
        )
        parser.add_argument('--dtemp', nargs='*', type=float, default=None,
                            help='Ensemble validations are made between two simulations at\n'
                                 'different state points.\n'
                                 'dtemp determines the temperature difference between base\n'
                                 'simulation and the additional point. If more than one\n'
                                 'value is given, several tests will be performed.\n'
                                 'By also giving dpress, both temperature and pressure can\n'
                                 'be displaced simultaneously.')
        parser.add_argument('--dpress', nargs='*', type=float, default=None,
                            help='Ensemble validations are made between two simulations at\n'
                                 'different state points.\n'
                                 'dpress determines the pressure difference between base\n'
                                 'simulation and the additional point. If more than one\n'
                                 'value is given, several tests will be performed.\n'
                                 'By also giving dtemp, both temperature and pressure can\n'
                                 'be displaced simultaneously.')
        parser.add_argument('-t', '--tolerance', type=float, default=3,
                            help=('The number of standard deviations a result can be off\n'
                                  'to be still accepted. Default: 3.'))

        return parser

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        args = cls.parser().parse_args(args)
        return cls.prepare(input_dir, target_dir, system_name, nobackup,
                           dtemp=args.dtemp, dpress=args.dpress)

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        args = cls.parser().parse_args(args)
        return cls.analyze(gmx_parser, system_dir, system_name, base_data, verbosity,
                           tolerance=args.tolerance, dtemp=args.dtemp, dpress=args.dpress)

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup, dtemp=None, dpress=None):
        # No standard values (system-dependent!)
        if not dtemp and not dpress:
            raise ValueError('Ensemble test for system ' + system_name +
                             ' has no defined temperature or pressure difference.')
        # Pad arrays - assume 0 difference if other difference is set
        if not dtemp:
            dtemp = [0] * len(dpress)
        if not dpress:
            dpress = [0] * len(dtemp)
        if len(dtemp) < len(dpress):
            dtemp.extend([0] * (len(dpress) - len(dtemp)))
        if len(dpress) < len(dtemp):
            dpress.extend([0] * (len(dtemp) - len(dpress)))
        # Check if we do any pressure differences
        no_press = True
        for dp in dpress:
            if dp*dp >= 1e-12:
                no_press = False
                break

        # Read base options
        options = GromacsInterface.read_mdp(os.path.join(input_dir, 'system.mdp'))

        # Check for target temperature
        if 'ref-t' in options:
            ref_t = float(options['ref-t'])
            ref_t_key = 'ref-t'
        elif 'ref_t' in options:
            ref_t = float(options['ref_t'])
            ref_t_key = 'ref_t'
        else:
            raise ValueError('Ensemble test for system ' + system_name + ' selected, ' +
                             'but system has no defined temperature.')

        # Check for target pressure, if relevant
        if no_press:
            ref_p = None
            ref_p_key = None
        elif 'ref-p' in options:
            ref_p = float(options['ref-p'])
            ref_p_key = 'ref-p'
        elif 'ref_p' in options:
            ref_p = float(options['ref_p'])
            ref_p_key = 'ref_p'
        else:
            raise ValueError('Ensemble test with pressure difference for system ' + system_name +
                             ' selected, but system has no defined pressure.')

        # Loop over tests
        directories = []
        for n, (dt, dp) in enumerate(zip(dtemp, dpress)):
            current_dir = os.path.join(target_dir, 'ensemble_' + str(n+1))
            mkdir_bk(current_dir, nobackup=nobackup)
            # change temperature & pressure
            options[ref_t_key] = str(ref_t + dt)
            if not no_press:
                options[ref_p_key] = str(ref_p + dp)
            # write mdp
            GromacsInterface.write_mdp(options, os.path.join(current_dir, 'system.mdp'))
            # copy topology and starting structure
            for suffix in ['gro', 'top']:
                shutil.copy2(os.path.join(input_dir, 'system.' + suffix),
                             current_dir)
            directories.append(current_dir)

        return directories

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity,
                tolerance=None, dtemp=None, dpress=None):
        # No standard values (system-dependent!)
        if not dtemp and not dpress:
            raise ValueError('Ensemble test for system ' + system_name +
                             ' has no defined temperature or pressure difference.')
        # Pad arrays - assume 0 difference if other difference is set
        if not dtemp:
            dtemp = [0] * len(dpress)
        if not dpress:
            dpress = [0] * len(dtemp)
        if len(dtemp) < len(dpress):
            dtemp.extend([0] * (len(dpress) - len(dtemp)))
        if len(dpress) < len(dtemp):
            dpress.extend([0] * (len(dtemp) - len(dpress)))

        nsystems = len(dtemp)

        if tolerance is None:
            tolerance = cls.parser().get_default('tolerance')

        # base data
        if base_data['reduced'] is None:
            current_dir = os.path.join(system_dir, 'base')
            base_data['reduced'] = gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.gro')
            )
        base_result = base_data['reduced']

        # list of results from simulations at different state points
        results = []
        for n in range(nsystems):
            current_dir = os.path.join(system_dir, 'ensemble_' + str(n+1))
            results.append(gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.gro')
            ))

        # run tests
        passed = True
        message = ''
        max_quantiles = -1
        for result, dt, dp in zip(results, dtemp, dpress):
            quantiles = ensemble.check(base_result, result, verbosity=verbosity)
            # filename=os.path.join(system_dir, system_name + '_ens'))
            if any(q > tolerance or math.isnan(q) for q in quantiles):
                passed = False
                if len(quantiles) == 1:
                    message += '\n    --dtemp={:.1f} --dpress={:.1f} : FAILED ({:.1f} quantiles off)'.format(
                        dt, dp, quantiles[0]
                    )
                else:
                    message += '\n    --dtemp={:.1f} --dpress={:.1f} : FAILED ([{:.1f}, {:.1f}] quantiles off)'.format(
                        dt, dp, quantiles[0], quantiles[1]
                    )
            else:
                if len(quantiles) == 1:
                    message += '\n    --dtemp={:.1f} --dpress={:.1f} : PASSED ({:.1f} quantiles off)'.format(
                        dt, dp, quantiles[0]
                    )
                else:
                    message += '\n    --dtemp={:.1f} --dpress={:.1f} : PASSED ([{:.1f}, {:.1f}] quantiles off)'.format(
                        dt, dp, quantiles[0], quantiles[1]
                    )
            max_quantiles = max(max_quantiles, max(quantiles))

        if passed:
            message = ('EnsembleTest PASSED (tolerance: {:.1f} quantiles)'.format(tolerance) +
                       message)
        else:
            message = ('EnsembleTest FAILED (tolerance: {:.1f} quantiles)'.format(tolerance) +
                       message)

        return {'test': passed,
                'result': max_quantiles,
                'tolerance': tolerance,
                'message': message}


class MaxwellBoltzmannTest(Test):
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(
            description='Tests the validity of the kinetic energy ensemble.',
            formatter_class=argparse.RawTextHelpFormatter,
            prog=cls.__name__
        )
        parser.add_argument('-t', '--tolerance', type=float, default=0.05,
                            help='The alpha value (confidence interval) used in\n'
                                 'the statistical test. Default: 0.05.')

        return parser

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        return cls.prepare(input_dir, target_dir, system_name, nobackup)

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        args = cls.parser().parse_args(args)
        return cls.analyze(gmx_parser, system_dir, system_name, base_data, verbosity,
                           alpha=args.tolerance)

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup):
        # no additional sims needed, base is enough
        # could check energy writing settings
        return []

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity, alpha=None):
        # Standard value
        if alpha is None:
            alpha = cls.parser().get_default('tolerance')

        # base data
        if base_data['reduced'] is None:
            current_dir = os.path.join(system_dir, 'base')
            base_data['reduced'] = gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.gro')
            )
        base_result = base_data['reduced']

        p = kinetic_energy.mb_ensemble(base_result, verbosity=verbosity)
        # filename=os.path.join(system_dir, system_name + '_mb'))

        if p >= alpha:
            message = 'MaxwellBoltzmannTest PASSED (p = {:g}, alpha = {:f})'.format(p, alpha)
        else:
            message = 'MaxwellBoltzmannTest FAILED (p = {:g}, alpha = {:f})'.format(p, alpha)

        return {'test': p >= alpha,
                'result': p,
                'tolerance': alpha,
                'message': message}


class EquipartitionTest(Test):
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(
            description='Tests the equipartition of the kinetic energy.',
            formatter_class=argparse.RawTextHelpFormatter,
            prog=cls.__name__
        )
        parser.add_argument('--distribution', default=False, action='store_true',
                            help=('If set, the groups of degrees of freedom will\n'
                                  'be separately tested as Maxwell-Boltzmann\n'
                                  'distributions. Otherwise, only the difference in\n'
                                  'average kinetic energy is checked.'))
        parser.add_argument('-t', '--tolerance', type=float, default=0.1,
                            help=('If --distribution is set:\n'
                                  '    The alpha value (confidence interval) used in the\n'
                                  '    statistical test. Default: 0.1.\n'
                                  'Else:\n'
                                  '    The maximum deviation allowed between groups.\n'
                                  '    Default: 0.1 (10%%)'))
        parser.add_argument('--random_groups', type=int, default=0,
                            help='Number of random division tests attempted.\n'
                                 'Default: 0 (random division tests off).')
        parser.add_argument('--random_divisions', type=int, default=2,
                            help='Number of groups the system is randomly divided in.\n'
                                 'Default: 2.')
        parser.add_argument('--molec_group', type=int, nargs='*', default=None, action='append',
                            help=('List of molecule indeces defining a group. Useful to\n'
                                  'pre-define groups of molecules\n'
                                  '(e.g. solute / solvent, liquid mixture species, ...).\n'
                                  'If not set, no pre-defined molecule groups will be\ntested.\n'
                                  'Note: If the last --molec-group flag is given empty,\n'
                                  'the remaining molecules are collected in this group.\n'
                                  'This allows, for example, to only specify the solute,\n'
                                  'and indicate the solvent by giving an empty flag.'))

        return parser

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup):
        # no additional sims needed, base is enough
        # could check position, velocity & energy writing settings
        return []

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        return cls.prepare(input_dir, target_dir, system_name, nobackup)

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        args = cls.parser().parse_args(args)
        return cls.analyze(gmx_parser, system_dir, system_name, base_data, verbosity,
                           tolerance=args.tolerance, distribution=args.distribution,
                           random_groups=args.random_groups, random_divisions=args.random_divisions,
                           molec_group=args.molec_group)

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity,
                tolerance=None, distribution=None,
                random_groups=None, random_divisions=None,
                molec_group=None):

        # Standard values
        if tolerance is None:
            tolerance = cls.parser().get_default('tolerance')
        if distribution is None:
            distribution = cls.parser().get_default('distribution')
        if random_groups is None:
            random_groups = cls.parser().get_default('random_groups')
        if random_divisions is None:
            random_divisions = cls.parser().get_default('random_divisions')
        if molec_group is None:
            molec_group = cls.parser().get_default('molec_group')

        # base data
        if base_data['full'] is None:
            current_dir = os.path.join(system_dir, 'base')
            base_data['full'] = gmx_parser.get_simulation_data(
                mdp=os.path.join(current_dir, 'mdout.mdp'),
                top=os.path.join(current_dir, 'system.top'),
                edr=os.path.join(current_dir, 'system.edr'),
                gro=os.path.join(current_dir, 'system.trr')
            )
        base_result = base_data['full']

        result = kinetic_energy.equipartition(base_result,
                                              dtemp=tolerance, alpha=tolerance,
                                              distribution=distribution, molec_groups=molec_group,
                                              random_divisions=random_divisions, random_groups=random_groups,
                                              verbosity=0)

        if distribution:
            res = min(result)
            test = res >= tolerance
            if test:
                message = 'EquipartitionTest PASSED (min p = {:g}, alpha = {:f})'.format(res, tolerance)
            else:
                message = 'EquipartitionTest FAILED (min p = {:g}, alpha = {:f})'.format(res, tolerance)
        else:
            dev_min = min(result)
            dev_max = max(result)
            if (1-dev_min) > (dev_max-1):
                res = dev_min
            else:
                res = dev_max
            test = res <= tolerance
            if test:
                message = 'EquipartitionTest PASSED (max dev = {:g}, tolerance = {:f})'.format(res, tolerance)
            else:
                message = 'EquipartitionTest FAILED (max dev = {:g}, tolerance = {:f})'.format(res, tolerance)

        return {'test': test,
                'result': res,
                'tolerance': tolerance,
                'message': message}


class KinConstraintsTest(Test):
    @classmethod
    def parser(cls):
        parser = argparse.ArgumentParser(
            description='Tests whether there is kinetic energy in constrained degrees of\nfreedom.',
            formatter_class=argparse.RawTextHelpFormatter,
            prog=cls.__name__
        )
        parser.add_argument('-t', '--tolerance', type=float, default=0.1,
                            help='Tolerance - TODO. Default: 0.1.')

        return parser

    @classmethod
    def prepare(cls, input_dir, target_dir, system_name, nobackup):
        # no additional sims needed, base is enough
        # could check if there are any constraints in the system
        return []

    @classmethod
    def prepare_parser(cls, input_dir, target_dir, system_name, nobackup, args):
        return cls.prepare(input_dir, target_dir, system_name, nobackup)

    @classmethod
    def analyze_parser(cls, gmx_parser, system_dir, system_name, base_data, verbosity, args):
        raise NotImplementedError

    @classmethod
    def analyze(cls, gmx_parser, system_dir, system_name, base_data, verbosity):
        raise NotImplementedError


all_tests = OrderedDict([
    ('integrator', IntegratorTest),
    ('ensemble', EnsembleTest),
    ('kin_mb', MaxwellBoltzmannTest),
    ('kin_equipartition', EquipartitionTest),
    ('kin_constraints', KinConstraintsTest)
])


def parse_systems(systems_json, systems_user, source_path,
                  analyze_only):
    # Parse json
    # As the order of the systems and the tests
    # might be meaningful, we need ordered dicts!
    system_list = json.load(systems_json)
    system_dict = OrderedDict()
    for system in system_list:
        system_name = system['name']
        system_dir = system['dir']
        # do the input files exist? (only relevant if we're not only analyzing)
        if not analyze_only:
            input_dir = os.path.join(source_path, system_dir, 'input')
            if not (os.path.isdir(input_dir) and
                    os.path.exists(os.path.join(input_dir, 'system.mdp')) and
                    os.path.exists(os.path.join(input_dir, 'system.top')) and
                    os.path.exists(os.path.join(input_dir, 'system.gro'))):
                raise ValueError('System ' + system_name + ' in ' +
                                 systems_json.name + ': Input files not found')
        # no need to run systems that we don't test
        if 'tests' not in system:
            raise ValueError('System ' + system_name + ' in ' +
                             systems_json.name + ' has no tests defined')
        test_list = system['tests']
        system['tests'] = OrderedDict()
        for test in test_list:
            test_name = test['test']
            if test_name not in all_tests:
                raise ValueError('Test ' + test_name + ' in ' +
                                 systems_json.name + ' is not a valid test')
            if test_name not in system['tests']:
                if 'args' in test:
                    test['args'] = [test['args'].split()]
                else:
                    test['args'] = [[]]
                system['tests'][test_name] = test
            else:
                if 'args' in test:
                    system['tests'][test_name]['args'].append(test['args'].split())
                else:
                    system['tests'][test_name]['args'].append([])

        system_dict[system_name] = system
        # add standard arguments
        if 'grompp_args' not in system:
            system['grompp_args'] = []
        else:
            system['grompp_args'] = system['grompp_args'].split()
        if 'mdrun_args' not in system:
            system['mdrun_args'] = []
        else:
            system['mdrun_args'] = system['mdrun_args'].split()

    # if user has not chosen otherwise, return full dict of systems
    if not systems_user:
        return system_dict

    # make sure user_dicts matches at least something
    for user_system in systems_user:
        for system in system_dict:
            if re.match(user_system + '$', system):
                break
        else:
            raise ValueError('System ' + user_system +
                             ' used in command line argument is not defined in ' +
                             systems_json.name)

    for system in list(system_dict.keys()):
        # delete systems not selected by user
        for user_system in systems_user:
            if re.match(user_system + '$', system):
                break
        else:
            system_dict.pop(system)

    # return reduced dict of systems
    return system_dict


def main(args):
    parser = argparse.ArgumentParser(
        description='Physical validation suite for GROMACS.',
        prog='gmx_physicalvalidation.py',
        formatter_class=argparse.RawTextHelpFormatter,
        epilog='Use --tests for details about the available tests and their arguments.'
    )
    parser.add_argument('json', type=open,
                        metavar='systems.json',
                        help='Json file containing systems and tests to be ran.')
    parser.add_argument('--tests', default=False, action='store_true',
                        help='Print details about the available tests and their arguments and exit.')
    parser.add_argument('-v', '--verbosity', type=int,
                        metavar='v', default=0,
                        help='Verbosity level. Default: 0 (quiet).')
    parser.add_argument('--gmx', type=str, metavar='exe', default=None,
                        help=('GROMACS executable. Default: Trying to use \'gmx\'.\n' +
                              'Note: If -p is used, the GROMACS executable is not needed.'))
    parser.add_argument('--bindir', type=str, metavar='dir', default=None,
                        help=('GROMACS binary directory.\n' +
                              'If set, trying to use \'bindir/gmx\' instead of plain \'gmx\'\n' +
                              'Note: If --gmx is set, --bindir and --suffix are ignored.'))
    parser.add_argument('--suffix', type=str, metavar='_s', default=None,
                        help=('Suffix of the GROMACS executable.\n' +
                              'If set, trying to use \'gmx_s\' instead of plain \'gmx\'\n' +
                              'Note: If --gmx is set, --bindir and --suffix are ignored.'))
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-p', '--prepare', action='store_true',
                       default=False,
                       help=('Only prepare simulations and output a \'run.sh\' file\n' +
                             'containing the necessary commands to run the systems.\n' +
                             'This allows to separate running simulations from analyzing them,\n' +
                             'useful e.g. to analyze the results on a different machine.\n' +
                             'Default: If none of \'-p\', \'-r\' or \'-a\' is given,\n' +
                             '         the systems are prepared, ran and analyzed in one call.'))
    group.add_argument('-r', '--run', action='store_true',
                       default=False,
                       help=('Only prepare and run simulations.\n' +
                             'Default: If none of \'-p\', \'-r\' or \'-a\' is given,\n' +
                             '         the systems are prepared, ran and analyzed in one call.'))
    group.add_argument('-a', '--analyze', action='store_true',
                       default=False,
                       help=('Only analyze previously ran simulations.\n' +
                             'This requires that the systems have been prepared using this program.\n' +
                             'Default: If none of \'-p\', \'-r\' or \'-a\' is given,\n' +
                             '         the systems are prepared, ran and analyzed in one call.'))
    parser.add_argument('-s', '--system', action='append', dest='systems',
                        metavar='system',
                        help=('Specify which system to run.\n' +
                              '\'system\' needs to match a system defined in the json file.\n' +
                              'Several systems can be specified by chaining several \'-s\' arguments.\n' +
                              'Defaults: If no \'-s\' argument is given, all systems and tests\n' +
                              '          defined in the json file are ran.\n' +
                              'Note: \'system\' can be a regular expression matching more than one system.'))
    parser.add_argument('--mpicmd', type=str, metavar='cmd', default=None,
                        help='MPI command used to invoke run command')
    parser.add_argument('--wd', '--working_dir', type=str,
                        metavar='dir', default=None,
                        help='Working directory (default: current directory)')
    parser.add_argument('--nobackup', default=False, action='store_true',
                        help='Do not create backups of files or folders.')
    
    if '--tests' in args:
        message = ('Physical validation suite for GROMACS\n'
                   'Available tests and their arguments\n'
                   '=====================================\n\n'
                   'The following tests can be specified in the json file:\n\n')
        for test in all_tests:
            message += '  * ' + test + '\n'

        message += '\nAll tests accept additional arguments:\n'
    
        for test, test_cls in all_tests.items():
            message += '\n'
            test_help = test_cls.parser().format_help()
            test_help = test_help.replace(test_cls.__name__, test).replace('usage: ', '')
            test_help = test_help.replace('  -h, --help            show this help message and exit\n', '')
            test_help = test_help.replace(' [-h]', '')
            test_help = test_help.replace(' ' * (len(test_cls.__name__) + 7), ' ' * len(test))
    
            first_line = test_help.split('\n')[0]
            separator = '-' * 70
            message += test_help.replace(first_line, separator + '\n' + first_line)

        sys.stderr.write(message)
        return

    args = parser.parse_args(args)

    # the input files are expected to be located where this script is
    source_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'systems')
    # target directory can be current or user chosen
    if args.wd is None:
        target_path = os.getcwd()
    else:
        if not os.path.exists(args.wd):
            os.makedirs(args.wd)
        target_path = args.wd

    # parse simulation stage to perform
    do_all = not (args.prepare or args.run or args.analyze)
    do_prepare = do_all or args.prepare or args.run
    write_script = args.prepare
    do_run = do_all or args.run
    do_analysis = do_all or args.analyze

    # get ordered dict of systems from combination of json file and user choices
    systems = parse_systems(args.json, args.systems, source_path, args.analyze)

    # prepare GROMACS interface
    if args.gmx:
        gmx = args.gmx
    else:
        gmx = 'gmx'
        if args.suffix:
            gmx += args.suffix
        if args.bindir:
            gmx = os.path.join(args.bindir, gmx)
    gmx_interface = None
    gmx_parser = None
    if do_run or do_analysis:
        gmx_interface = GromacsInterface(exe=gmx)
        gmx_parser = GromacsParser(exe=gmx)

    if do_prepare:
        nsystems = len(systems)
        n = 0
        runs = []  # this will contain all information needed to run the system
        for system_name, system in systems.items():
            n += 1
            print('\rPreparing run files for systems... [{:d}/{:d}] '.format(n, nsystems), end='')
            sys.stdout.flush()  # py2 compatibility
            system_dir = system['dir']
            system_dirs = []  # list of directories with subsystems
            # prepare the base system
            input_dir = os.path.join(source_path, system_dir, 'input')
            target_dir = os.path.join(target_path, system_dir)
            mkdir_bk(target_dir, nobackup=args.nobackup)
            basedir = os.path.join(target_dir, 'base')
            mkdir_bk(basedir, nobackup=args.nobackup)
            for suffix in ['mdp', 'gro', 'top']:
                shutil.copy2(os.path.join(input_dir, 'system.' + suffix),
                             basedir)
            system_dirs.append(basedir)
            # call prepare method of chosen tests
            for test_name, test in system['tests'].items():
                for test_args in test['args']:
                    system_dirs.extend(
                        all_tests[test_name].prepare_parser(input_dir, target_dir, system_name,
                                                            args.nobackup, test_args)
                    )

            # save run information
            for d in system_dirs:
                runs.append({
                    'dir': d,
                    'grompp_args': system['grompp_args'],
                    'mdrun_args': system['mdrun_args']
                })
        # end of loop over systems
        print('-- done.')

        if write_script:
            print('Writing run script... ', end='')
            sys.stdout.flush()  # py2 compatibility
            script_file = os.path.join(target_path, 'run_simulations.sh')
            if not args.nobackup:
                file_bk(script_file)
            with open(script_file, 'w') as f:
                f.write('# This file was created by the physical validation suite for GROMACS.\n')
                f.write('\n# Define run variables\n')
                f.write('WORKDIR=' + os.path.abspath(target_path) + '\n')
                f.write('GROMPPCMD="' + os.path.abspath(gmx) + ' grompp"\n')
                f.write('MDRUNCMD="' + os.path.abspath(gmx) + ' mdrun"\n')
                f.write('\n# Run systems\n')
                f.write('startpath=$PWD\n')
                f.write('cd $WORKDIR\n')
                for run in runs:
                    for cmd in basic_run_cmds(directory=os.path.relpath(os.path.abspath(run['dir']),
                                                                        os.path.abspath(target_path)),
                                              grompp_args=run['grompp_args'],
                                              mdrun_args=run['mdrun_args']):
                        f.write(cmd + '\n')
                    f.write('\n')
                f.write('cd $startpath\n')
            print('-- done.')
            print('Run script written to ' + script_file)
            print('Adapt script as necessary and run simulations. Make sure to preserve the folder structure!')
            print('Once all simulations have ran, analyze the results using `make check-phys-analyze` or '
                  'using the `-a` flag of `gmx_physicalvalidation.py`.')
        # end if write_script

        if do_run:
            nruns = len(runs)
            # send messages from GROMACS to log
            gmx_log = open(os.path.join(target_path, 'physicalvalidation_gmx.log'), 'w')
            for n, run in enumerate(runs):
                print('\rRunning (sub)systems... [{:d}/{:d}] '.format(n+1, nruns), end='')
                sys.stdout.flush()  # py2 compatibility
                gmx_interface.grompp(mdp='system.mdp',
                                     top='system.top',
                                     gro='system.gro',
                                     tpr='system.tpr',
                                     cwd=run['dir'],
                                     args=run['grompp_args'],
                                     stdout=gmx_log,
                                     stderr=gmx_log)
                gmx_interface.mdrun(tpr='system.tpr',
                                    deffnm='system',
                                    cwd=run['dir'],
                                    args=run['mdrun_args'],
                                    stdout=gmx_log,
                                    stderr=gmx_log,
                                    mpicmd=args.mpicmd)
            gmx_log.close()
            print('-- done.')
        # end if do_run
    # end if do_prepare

    if do_analysis:
        title = 'GROMACS PHYSICAL VALIDATION RESULTS'
        width = 70
        indent = int((width - len(title))/2)
        print()
        print(' ' * indent + '=' * len(title))
        print(' ' * indent + title)
        print(' ' * indent + '=' * len(title))
        print()
        passed = True
        for system_name, system in systems.items():
            system_dir = system['dir']
            # save system data if re-used for different test
            # massively reduces run time of multiple tests
            system_data = {
                'reduced': None,
                'full': None
            }
            # system directory
            target_dir = os.path.join(target_path, system_dir)

            print('Analyzing system ' + system_name)

            # call analyze method of chosen tests
            for test_name, test in system['tests'].items():
                for test_args in test['args']:
                    try:
                        result = all_tests[test_name].analyze_parser(gmx_parser, target_dir,
                                                                     system_name, system_data,
                                                                     args.verbosity, test_args)
                    except Exception as err:
                        print('    ' + all_tests[test_name].__name__ + ' FAILED (Exception in evaluation)')
                        print('    '*2 + type(err).__name__ + ': ' + str(err))
                        passed = False
                    else:
                        for line in result['message'].split('\n'):
                            print('    ' + line)

                        passed = passed and result['test']
            # end loop over tests
            print()
        # end loop over systems
        return int(not passed)
    # end if do_analysis

    # assuming everything is ok if we ended up here
    return 0


if __name__ == "__main__":
    return_value = main(sys.argv[1:])
    sys.exit(return_value)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""
Abstract base classes for gmxapi.simulation module
==================================================

These classes allow static and dynamic type checking for gmxapi Python
interfaces. Some may be used as bases to inherit behaviors, but users should
not assume that a gmxapi object actually inherits from classes in this module.

For more information on the concept of abstract base classes in Python, refer
to https://docs.python.org/3/library/abc.html

For more on type hinting, see https://docs.python.org/3/library/typing.html
"""

import abc


class ModuleObject(abc.ABC):
    """Extended interface for objects in the gmaxpi.simulation module.

    Implies availability of additional binding details.
    """



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structures carrying simulation data.
"""
from ..util import error as pv_error
# py2.7 compatibility
from .unit_data import UnitData
from .ensemble_data import EnsembleData
from .system_data import SystemData
from .observable_data import ObservableData
from .trajectory_data import TrajectoryData
# replace lines above by this when py2.7 support is dropped:
# from . import UnitData, EnsembleData, SystemData, ObservableData, TrajectoryData


class SimulationData(object):
    r"""SimulationData: System information and simulation results

    The SimulationData class holds both the information on the system and
    the results of a simulation run of that system. SimulationData contains
    all information on a simulation run needed by the physical validation
    tests. SimulationData objects can either be created directly by calling
    the class constructor, or by using a parser returning a SimulationData
    object.
    """

    @staticmethod
    def compatible(data_1, data_2):
        r"""Checks whether two simulations are compatible for common validation.

        Parameters
        ----------
        data_1 : SimulationData
        data_2 : SimulationData

        Returns
        -------
        result : bool

        """
        if not isinstance(data_1, SimulationData):
            raise pv_error.InputError('data_1',
                                      'Expected type SimulationData')
        if not isinstance(data_2, SimulationData):
            raise pv_error.InputError('data_2',
                                      'Expected type SimulationData')

        return data_1.units == data_2.units

    def __init__(self, units=None, dt=None,
                 system=None, ensemble=None,
                 observables=None, trajectory=None):
        self.__units = None
        if units is not None:
            self.units = units
        self.__dt = 0
        if dt is not None:
            self.dt = dt
        self.__system = None
        if system is not None:
            self.system = system
        self.__ensemble = None
        if ensemble is not None:
            self.ensemble = ensemble
        self.__observables = None
        if observables is not None:
            self.observables = observables
        self.__trajectory = None
        if trajectory is not None:
            self.trajectory = trajectory

    @property
    def ensemble(self):
        r"""EnsembleData: Information on the sampled ensemble

        Returns
        -------
        ensemble : EnsembleData
        """
        return self.__ensemble

    @ensemble.setter
    def ensemble(self, ensemble):
        if not isinstance(ensemble, EnsembleData):
            raise TypeError('No known conversion from ' + str(type(ensemble)) +
                            'to EnsembleData')
        self.__ensemble = ensemble

    @property
    def units(self):
        r"""UnitsData: Information on the sampled units

        Returns
        -------
        units : UnitData
        """
        return self.__units

    @units.setter
    def units(self, units):
        if not isinstance(units, UnitData):
            raise TypeError('No known conversion from ' + str(type(units)) +
                            'to UnitData')
        self.__units = units

    @property
    def observables(self):
        r"""ObservableData: Observables collected during the simulation

        Returns
        -------
        observables : ObservableData
        """
        return self.__observables

    @observables.setter
    def observables(self, observables):
        if not isinstance(observables, ObservableData):
            raise TypeError('No known conversion from ' + str(type(observables)) +
                            'to ObservableData')
        self.__observables = observables

    @property
    def trajectory(self):
        r"""TrajectoryData: Trajectories collected during the simulation

        Returns
        -------
        trajectory : TrajectoryData
        """
        return self.__trajectory

    @trajectory.setter
    def trajectory(self, trajectory):
        if not isinstance(trajectory, TrajectoryData):
            raise TypeError('No known conversion from ' + str(type(trajectory)) +
                            'to TrajectoryData')
        self.__trajectory = trajectory

    @property
    def system(self):
        r"""SystemData: Information on the system's system

        Returns
        -------
        system : SystemData
        """
        return self.__system

    @system.setter
    def system(self, system):
        if not isinstance(system, SystemData):
            raise TypeError('No known conversion from ' + str(type(system)) +
                            'to SystemData')
        self.__system = system

    @property
    def dt(self):
        r""" The timestep of the simulation run.

        Returns
        -------
        timestep : float
        """
        return self.__dt

    @dt.setter
    def dt(self, dt):
        dt = float(dt)
        self.__dt = dt

    def set_ensemble(self, ensemble,
                     natoms=None, mu=None,
                     volume=None, pressure=None,
                     energy=None, temperature=None):
        self.__ensemble = EnsembleData(ensemble,
                                       natoms, mu,
                                       volume, pressure,
                                       energy, temperature)



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2022- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Script to dispatch and run release pipeline.

When run locally, submits a build pipeline to the
GitLab server with the necessary information to later
automatically upload generated artefacts to the relevant servers.

The script has two modes, local mode submits the job pipeline, while
in server mode we sync artefacts with the server.
Requires the `python_gitlab <https://python-gitlab.readthedocs.io/en/stable/index.html>`__
package for Python access to the GitLab API.

Author:
    * Andrey Alekseenko <al42and@gmail.com>
    * Mark Abraham <mark.j.abraham@gmail.com>
    * Paul Bauer <paul.bauer.q@gmail.com>
    * Eric Irrgang <ericirrgang@gmail.com>
"""

import argparse
from pathlib import Path
import os
import re
from shutil import copyfile
import subprocess
import typing


def submit_gitlab_pipeline(
    auth_token, ssh_key, branch="main", release_build=False, dry_run=True
):
    """Submit a pipeline to GitLab server to run a release build.

    Authenticates user using token to the project, and tries to create new GROMACS_RELEASE
    pipeline. If release_build is true, and dry_run is false, this will generate a pipeline
    that will upload real artefacts to the manual and ftp servers if run from a release branch
    (or main).  Otherwise, the generated pipeline will always only upload files to a
    test location.

    The pipeline is by default created for main, but can be run for any branch.

    Needs access to gitlab python API. For this, someone running the script needs
    to have installed the python library (pip install python-gitlab) and created
    a personal access token that grants access to the GROMACS project
    (https://docs.gitlab.com/ee/user/profile/personal_access_tokens.html).

    Throws if the authentication process, or the submission of the pipeline fails.
    """
    import gitlab

    gl = gitlab.Gitlab("https://gitlab.com", private_token=auth_token)
    # The project ID for GROMACS is hardcoded here
    project_id = 17679574
    project = gl.projects.get(project_id)

    # Set some variables
    release_build_str = f"{release_build}".lower()
    dry_run_str = f"{dry_run}".lower()

    ssh_key_string = Path(ssh_key).read_text()

    # add trailing new line for the ssh-key
    if ssh_key_string[:-1] != "\n":
        ssh_key_string += "\n"

    print("Going to start pipeline with following arguments")
    print(r"Branch = ", branch)
    print(r"GROMACS_RELEASE = ", release_build_str)
    print(r"DRY_RUN = ", dry_run_str)
    # print(r'SSH_PRIVATE_KEY = ', ssh_key_string)

    pipeline = project.pipelines.create(
        {
            "ref": branch,
            "variables": [
                {"key": "GROMACS_RELEASE", "value": release_build_str},
                {"key": "SSH_PRIVATE_KEY", "value": ssh_key_string},
                {"key": "DRY_RUN", "value": dry_run_str},
            ],
        }
    )


def upload_files(
    *,
    path: typing.Union[str, Path],
    options: typing.List[str],
    server: str,
    dry_run=False,
):
    """Actual upload command.

    Takes care of moving files to the server for public consumption.
    Needs to have successfully set up SSH key infrastructure before.
    """

    upload_command = ["rsync", "-rlvP"]
    upload_command.extend(options)

    upload_command.append(str(path))

    upload_command.append(server)

    print(upload_command)

    if not dry_run:
        ret = subprocess.run(upload_command, capture_output=True)
        if ret.returncode != 0:
            print(ret.stdout)
            print(ret.stderr)
            exit(ret.returncode)


def upload_release_artifacts():
    """Upload files to server.

    Sets up infrastructure for ssh to be able to authenticate to server and upload files.
    Key and host information is read from the GitLab environment variables that hold them
    after submitting the job.

    The following need to be true for uploading files to the actual location on
    the GROMACS project ftp and webservers:
        * Pipeline is run for either main, or one of the release branches
        * The pipeline is invoked with "GROMACS_RELEASE" set to "true"
        * The "DRY_RUN" variable needs to be explicitly set to "false"

    If any of those conditions above is not met, files are instead uploaded to a test
    location on the server (path prefixed with ".ci-test"), or, if "DRY_RUN" is not "false",
    no upload is attempted.

    For uploading a new version, the script will never overwrite existing files, as
    an additional measure of ensuing no problems with existing, world visible, files.
    """

    # Please see the variable documentation in docs/dev-manual/gitlab-ci.rst for explanation
    # of GROMACS specific  variables.
    dry_run_str = os.getenv("DRY_RUN")
    release_build_str = os.getenv("GROMACS_RELEASE")
    branch = os.getenv("CI_COMMIT_BRANCH")
    build_dir = os.getenv("BUILD_DIR")
    version = os.getenv("VERSION")

    if dry_run_str.lower() not in ("true", "false"):
        raise ValueError(
            f'Wrong value of DRY_RUN: "{dry_run_str}". Only "true" and "false" are allowed'
        )

    # we keep track of whether any command has failed so far
    is_upload = False
    dry_run = True
    if dry_run_str == "false":
        dry_run = False

    full_version = version
    if not release_build_str == "true":
        full_version += "-dev"

    overwrite_str = ""
    # the default location for ftp and manual bot is ./, so we just add .ci-test to upload to the test location
    upload_location = "./.ci-test/"
    if re.match(r"^release-\d{4}$", branch) or re.match(r"^main$", branch):
        if release_build_str == "true":
            is_upload = True
            overwrite_str = "--ignore-existing"
            upload_location = "./"
            # Only upload to real location if all preconditions are set. As the default path is as mentioned above,
            # we just use the directory above it

    current_dir = os.getcwd()

    # set up manual front page repo
    os.mkdir("manual-front-page")
    os.chdir("manual-front-page")
    ret_init = subprocess.run(["git", "init"], capture_output=True)
    ret_fetch = subprocess.run(
        [
            "git",
            "fetch",
            "https://gitlab.com/gromacs/deployment/manual-front-page.git",
            "main",
        ],
        capture_output=True,
    )
    ret_checkout = subprocess.run(
        ["git", "checkout", "-qf", "FETCH_HEAD"], capture_output=True
    )
    os.chdir(current_dir)
    if ret_init.returncode != 0:
        print(ret_init.stdout)
        print(ret_init.stderr)
        exit(1)
    if ret_fetch.returncode != 0:
        print(ret_fetch.stdout)
        print(ret_fetch.stderr)
        exit(1)
    if ret_checkout.returncode != 0:
        print(ret_checkout.stdout)
        print(ret_checkout.stderr)
        exit(1)

    ret = subprocess.run(
        ["make", "html"], capture_output=True, cwd=Path.cwd() / "manual-front-page"
    )

    if ret.returncode != 0:
        print(ret.stdout)
        print(ret.stderr)
        exit(1)

    ftp_server = "ftpbot@ftp.gromacs.org"
    manual_server = "manualbot@manual.gromacs.org"

    website_file_location = build_dir + "/docs/html"
    frontpage_file_location = "manual-front-page/_build/html"
    manual_file = f"manual-{version}.pdf"
    manual_file_location = website_file_location + "/" + manual_file

    website_path_on_server = f"{manual_server}:{upload_location}/{full_version}/"
    source_path_on_server = f"{ftp_server}:{upload_location}/gromacs/"
    regressiontests_path_on_server = f"{ftp_server}:{upload_location}/regressiontests/"
    manual_path_on_server = f"{ftp_server}:{upload_location}/manual/"
    frontpage_path_on_server = f"{manual_server}:{upload_location}/"

    website_upload_options = ["--chmod=u+rwX,g+rwX,o+rX"]
    file_upload_options = ["--chmod=u+rw,g+rw,o+r"]

    frontpage_upload_options = website_upload_options + [
        "--exclude",
        "_sources",
        "--exclude",
        ".buildinfo",
        "--exclude",
        "objects.inv",
    ]

    if is_upload and overwrite_str:
        website_upload_options.append(overwrite_str)
        file_upload_options.append(overwrite_str)

    manual_file = f"manual-{full_version}.pdf"
    copyfile(manual_file_location, manual_file)
    source_tarball = f"gromacs-{full_version}.tar.gz"
    regressiontests_tarball = f"regressiontests-{full_version}.tar.gz"

    os.chdir(website_file_location)
    upload_files(
        path="./",
        options=website_upload_options,
        server=website_path_on_server,
        dry_run=dry_run,
    )
    os.chdir(current_dir)
    upload_files(
        path=manual_file,
        options=file_upload_options,
        server=manual_path_on_server,
        dry_run=dry_run,
    )
    upload_files(
        path=source_tarball,
        options=file_upload_options,
        server=source_path_on_server,
        dry_run=dry_run,
    )
    upload_files(
        path=regressiontests_tarball,
        options=file_upload_options,
        server=regressiontests_path_on_server,
        dry_run=dry_run,
    )
    os.chdir(frontpage_file_location)
    upload_files(
        path="./",
        options=frontpage_upload_options,
        server=frontpage_path_on_server,
        dry_run=dry_run,
    )
    os.chdir(current_dir)


parser = argparse.ArgumentParser(description="Automatic release options.")

mode_group = parser.add_mutually_exclusive_group(required=True)
mode_group.add_argument(
    "--local",
    action="store_true",
    help="Set when running in local (submit pipeline) mode.",
)
mode_group.add_argument(
    "--server",
    action="store_const",
    const=False,
    dest="local",
    help="Set when running in server (upload artefacts) mode.",
)

parser.add_argument(
    "--token", type=str, help="GitLab access token needed to launch pipelines"
)
parser.add_argument(
    "--ssh-key",
    type=str,
    help="Path to SSH key needed to upload things to server. Pass in local mode to have it during the job",
)

release_parser = parser.add_mutually_exclusive_group(required=False)
release_parser.add_argument("--release", dest="release", action="store_true")
release_parser.add_argument(
    "--no-release", dest="release", action="store_const", const=False
)

dry_run_parser = parser.add_mutually_exclusive_group(required=False)
dry_run_parser.add_argument("--dry-run", dest="dry_run", action="store_true")
dry_run_parser.add_argument(
    "--no-dry-run", dest="dry_run", action="store_const", const=False
)

parser.add_argument(
    "--branch",
    type=str,
    default="main",
    help='Branch to run pipeline for (default "main")',
)


if __name__ == "__main__":

    args = parser.parse_args()

    if args.local:
        if (
            args.token is None
            or args.branch is None
            or args.release is None
            or args.dry_run is None
            or args.ssh_key is None
        ):
            raise RuntimeError(
                "Need to provide all command line options for running in local mode"
            )

        ssh_key_location = Path(args.ssh_key).resolve()

        submit_gitlab_pipeline(
            args.token, ssh_key_location, args.branch, args.release, args.dry_run
        )

    else:
        # --server mode
        upload_release_artifacts()
